use std::collections::BTreeMap;

use rand::thread_rng;
use rand::seq::SliceRandom;

use actix::prelude::*;

use crate::server::server::{GameConnection, Server};
use crate::datasource;
use crate::game;
use crate::websocket;
use crate::websocket::*;

use crate::server::messages::*;

#[derive(PartialEq)]
enum LobbyState {
    Starting,
    Creating,
    Guessing,
    Revealing
}

pub struct GameLobby {
    connected_players: BTreeMap<String, Addr<GameConnection>>,
    game_id: String,
    game: game::Game,
    waiting_players: BTreeMap<String, Addr<GameConnection>>,
    ready_players: Vec<String>,
    lobby_state: LobbyState,
    data_source: Box<dyn datasource::DataSource<String>>,
}

impl Actor for GameLobby {
    type Context = Context<Self>;
    fn started(&mut self, ctx: &mut Self::Context) {
        self.lobby_state = LobbyState::Starting;
    }
}

impl Handler<JoinRequest> for GameLobby {
    type Result = Answer;
    fn handle(&mut self, jr: JoinRequest, ctx: &mut Self::Context) -> Self::Result {
        if self.lobby_state == LobbyState::Starting {
            jr.p.do_send(LobbyJoined(ctx.address()));
            self.connected_players.insert(jr.nick.clone(), jr.p);
            self.game.player_join(jr.nick);
            self.send_game_to_all();
            Answer::LobbyJoined(ctx.address())
        }
        else {
            self.waiting_players.insert(jr.nick.clone(), jr.p);
            Answer::NoSuchLobby
        }
    }
}

impl Handler<ReadyMsg> for GameLobby {
    type Result = ();
    fn handle(&mut self, rm: ReadyMsg, ctx: &mut Self::Context) -> Self::Result {
        if !self.ready_players.contains(&rm.0) {
            self.ready_players.push(rm.0);
        }
        if self.ready_players.len() >= self.connected_players.len() {
            self.set_state(LobbyState::Creating);
            self.send_game_to_all();
        }
    }
}

impl Handler<SubmitWordMsg> for GameLobby {
    type Result = ();
    fn handle(&mut self, swm: SubmitWordMsg, ctx: &mut Self::Context) -> Self::Result {
        let correct = self.game.submit_creation(&swm.nick, swm.word);
        if self.game.all_words_submitted() {
            self.set_state(LobbyState::Guessing);
            self.game.next_state();
            self.send_game_to_all();
        }
    }
}

impl Handler<SubmitGuessMsg> for GameLobby {
    type Result = ();
    fn handle(&mut self, sgm: SubmitGuessMsg, ctx: &mut Self::Context) -> Self::Result {
        self.game.submit_guess(&sgm.nick, sgm.guesses);
        if self.game.all_guesses_submitted() {
            self.set_state(LobbyState::Revealing);
            let results = self.create_result_data();
            self.broadcast_results(results);

            self.game.next_state();
            self.send_game_to_all();
        }
    }
}

impl GameLobby {
    pub fn new(gi: String) -> Self {
        GameLobby {
            connected_players: BTreeMap::new(),
            game_id: gi,
            game: game::Game::new(),
            waiting_players: BTreeMap::new(),
            ready_players: Vec::new(),
            lobby_state: LobbyState::Starting,
            data_source: Box::new(datasource::ArraySource::create(vec![
                "Delikatessfutter für Hunde".to_owned(),
                "Ein Hotel für die ganze Familie".to_owned(),
                "Brasilianischer Superstar".to_owned(),
                "Buchstabe des griechischen Alphabets".to_owned(),
            ]))
        }
    }

    fn set_state(&mut self, new_state: LobbyState) {

        match new_state {
            LobbyState::Starting => {
                for (nick, _addr) in &self.waiting_players {
                    self.game.player_join(nick.clone());
                }
                self.connected_players.append(&mut self.waiting_players);
                self.ready_players.clear();
            },
            LobbyState::Creating => {
                let ds = &self.data_source;
                let mut index = 0;
                self.game.start_round(|| {
                    index += 1;
                    ds.get_ith(index - 1).unwrap().clone()
                });
            },
            _ => {}
        }
        self.lobby_state = new_state;
    }

    pub fn send_game_to_all(&self) {
        for (nick, player) in &self.connected_players {
            let game_state = self.create_opaque_message(nick);
            player.do_send(GameUpdate{ game_data: game_state });
        }
    }

    pub fn create_result_data(&self) -> RoundResultData {
        let results_table = self.game.create_results();

        let words = self.game.players.iter()
                        .map(|p| p.submitted_word.clone().unwrap())
                        .collect::<Vec<_>>();
        let questions = self.game.players.iter()
                        .map(|x| x.creating_exercise.as_ref().unwrap().question.clone())
                        .chain(self.game.additional_questions.clone().into_iter())
                        .collect::<Vec<_>>();
        let solutions = words.iter()
                        .map(|x| x.clone())
                        .zip(questions.iter().map(|x| x.clone()))
                        .collect::<Vec<_>>();

        let guesses = self.game.create_results();

        RoundResultData {
            words,
            questions,
            solutions,
            guesses
        }
    }

    pub fn broadcast_results(&self, results: RoundResultData) {
        for (_nick, player) in &self.connected_players {
            player.do_send(ResultMsg{ results: results.clone() });
        }
    }

    pub fn create_opaque_message(&self, nick: &str) -> GameData {

        let player = self.game.players.iter()
                        .find(|p| p.nick == nick);
        GameData {
            players: self.game.players.iter()
                        .map(|p| websocket::PlayerData{ nick: p.nick.clone(), points: 0 })
                        .collect::<Vec<_>>(),
            state_data:
                match self.game.state {
                    game::GameState::Starting => {
                        GameStateData::Starting
                    },
                    game::GameState::Creating => {
                        GameStateData::Creating{
                            question: player.unwrap().creating_exercise.as_ref().unwrap().question.clone(),
                            available_chars: player.unwrap().creating_exercise.as_ref().unwrap().letters.clone()
                        }
                    },
                    game::GameState::Guessing => {
                        self.create_guessing()
                    },
                }
        }
    }

    pub fn check_optionals(&self) {
        for p in &self.game.players {
            println!("{:?}", p);
        }
    }

    pub fn create_guessing(&self) -> GameStateData {
        self.check_optionals();
        let words_with_chars = self.game.players.iter()
            .map(|p|
                (p.submitted_word.clone().unwrap(),
                p.creating_exercise.clone().unwrap().letters))
            .collect::<Vec<_>>();

        let mut questions = self.game.players.iter()
            .map(|p| p.creating_exercise.clone().unwrap().question)
            .collect::<Vec<_>>();
        questions.append(&mut self.game.additional_questions.iter()
                            .map(|x| x.clone())
                            .collect::<Vec<_>>()
                        );

        questions.shuffle(&mut thread_rng());

        GameStateData::Guessing {
            submitted_words: words_with_chars,
            questions: questions
        }
    }
}